#!/usr/bin/perl -w

###############################################################
#
#      File: SayWhat::Tk.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: Tk interface for SayWhat
# 
###############################################################

package SayWhat::Tk;
use Tk;
use Tk::FileSelect;
use Tk::Pane;
use Speech::Festival::Synthesiser;
use base qw(SayWhat::Base);

our $Tree_Font = 'helvetica -10 bold';
our $Tree_SkipX = 10;
our $Tree_SkipY = 30;
our $Tree_LinePad = 1;
our $Tree_BoxActive = 0;

our ($padx,$pady) = (2,2);
our $BUTTONS_PER_ROW = 2;

sub new {
  my ($that,%args) = @_;
  my $self = $that->SUPER::new(%args);
  $self->initialize(%$args);
  return $self;
}

sub initialize {
  my ($self,%args) = @_;

  my $tk = $self->{tk} = {main => MainWindow->new()};
  my $main = $tk->{main};
  $main->title('TkSayWhat');

  #------------------------------------------------------------
  # file selector
  #------------------------------------------------------------
  my $fs = $tk->{file_selector} = $main->FileSelect();
  $fs->Subwidget(file_entry)->bind('<Key-Return>', sub { file_entry_return_click($fs); });

  my $lframe = $tk->{lframe} = $main->Frame();
  my $rframe = $tk->{rframe} = $main->Frame();

  #------------------------------------------------------------
  # expanding frame
  #------------------------------------------------------------
  my $xframe = $tk->{xframe} = $lframe->Frame();
  my $text_label = $tk->{text_label} = $xframe->Label(-text => 'Utterance string:');
  my $text_field = $tk->{text_field} = $xframe->Scrolled('Text',
							 -relief => 'sunken',
							 -borderwidth => 2,
							 -scrollbars => 'e',
							 #-width => 20,
							 -height => 10,
							 -background => 'white'
							);

  $text_label->pack(-side => 'top', -fill => 'x', -padx => $padx);
  $text_field->pack(-side => 'top',  -fill => 'both', -expand => 1, -padx => $padx);

  #------------------------------------------------------------
  # generation / grammar frame
  #------------------------------------------------------------
  my $rrframe = $tk->{rrframe} = $rframe->Frame();
  my $gen_frame = $tk->{gen_frame} = $rrframe->Frame();

  ## rename-utterance field
  $tk->{new_uttid} = '';
  my $rename_label = $tk->{rename_label} = $gen_frame->Label(-text => 'Utterance Name:');
  my $rename_entry = $tk->{rename_entry} = $gen_frame->Entry(-textvariable => \$tk->{new_uttid},
							     -background => 'white',
							     -width => 10
							    );
  ## generation-root nonterm
  $tk->{gen_nonterm} = 'S';
  my $gen_label = $tk->{gen_label} = $gen_frame->Label(-text => 'Generation Root:');
  my $gen_entry = $tk->{gen_entry} = $gen_frame->Entry(-textvariable => \$tk->{gen_nonterm},
						       -background => 'white',
						       -width => 10
						      );
  ## word-boundary
  my $wb_label = $tk->{wb_label} = $gen_frame->Label(-text => 'Word Boundary:');
  my $wb_entry = $tk->{wb_entry} = $gen_frame->Entry(-textvariable => \$self->{wbound},
						     -background => 'white',
						     -width => 10
						    );

  #$gen_entry->bind('<Enter>' => \&SayWhat::Tk::generate_click);

  $rename_label->pack(-side => 'top');
  $rename_entry->pack(-side => 'top', -fill => 'x', -expand => 1);
  $rename_entry->bind('<Key-Return>' => sub { $self->rename_click; });

  $gen_label->pack(-side => 'top');
  $gen_entry->pack(-side => 'top', -fill => 'x');

  $wb_label->pack(-side => 'top');
  $wb_entry->pack(-side => 'top', -fill => 'x', -expand => 1);
  #$wb_entry->bind('<Key-Return>' =>
	#	  sub {
	#	    #$self->{wbound} = $wbound_str;
	#	    $tk->{message} = "Set word-boundary to `$self->{wbound}'.";
	#	  });


  #------------------------------------------------------------
  # Listbox frame
  #------------------------------------------------------------
  my $list_label = $tk->{list_label} = $rrframe->Label(-text => 'Known Utterances:');
  my $listbox = $tk->{listbox} = $rrframe->Scrolled('Listbox',
						    -scrollbars => 'se',
						    -background => 'white',
						    -selectmode => 'browse',
						    -exportselection => 0,
						   );
  #$listbox->bind("<Button-1>", sub { show_tree($self) });
  $listbox->bind("<Button-1>", sub { $self->update });

  #------------------------------------------------------------
  # Buttons Frame
  #------------------------------------------------------------
  #my $buttons_frame = $tk->{buttons_frame} = $rframe->Frame(-borderwidth => 2, -relief => 'raised');
  #my $buttons_label = $tk->{buttons_frame} = $buttons_frame->Label(-text => 'Operations:');
  #$buttons_label->pack(-side => 'top', -fill => 'x');

  # --- OBSOLETE ---
  #my $voice_sel = $tk->{voice_sel} = $buttons_frame->Menubutton(-text => 'Voice',
	#							 -relief => 'raised',
	#							 -indicatoron => 1,
	#							 -width => 12
	#							);
  #$voice_sel->pack(-side => 'top', -fill => 'x');

  #my $style_sel = $tk->{style_sel} = $buttons_frame->Menubutton(-text => 'Style',
	#							 -relief => 'raised',
	#							 -indicatoron => 1,
	#							 -width => 12
	#							);
  #$style_sel->pack(-side => 'top', -fill => 'x');


  my $buttons_frame = $tk->{buttons_frame} = $tk->{main}->Frame();

  #------------------------------------------------------------
  # Buttons
  #   $button = $self->new_button($frame_name,$name,$text,$command);
  #------------------------------------------------------------

  ## add/delete
  # $frame = $self->new_button_frame($parent_name,$name,$label_text,@button_specs);
  #  + where each $button_spec in @button_specs is a listref:
  #    [ $name, $text, $cmd, %args ]
  $self->new_button_frame('buttons_frame',0,0,$BUTTONS_PER_ROW,'maintain_frame','Utterance Maintainance',
			  ['add_button','Add',sub { $self->add_click(); }],
			  ['del_button','Delete',sub { $self->del_click(); }],
			 );

  ## generate/gensay
  $self->new_button_frame('buttons_frame',0,1,$BUTTONS_PER_ROW,'generate_frame','Generation',
			  ['generate_button','Generate',sub { $self->generate_click(); }],
			  ['generate_say_button','Generate & Say',sub { $self->generate_say_click(); }],
			 );

  ## grammar stuff
  $self->new_button_frame('buttons_frame',1,0,$BUTTONS_PER_ROW,'grammar_frame','Grammar Operations',
			  ['load_grammar_button','Load Grammar',sub { $self->load_grammar_click(); }],
			  ['retrieve_grammar_button','Retrieve Grammar',sub { $self->retrieve_grammar_click(); }],
			  ['store_grammar_button','Store Grammar',sub { $self->store_grammar_click(); }],
			  ['forget_grammar_button','Forget Grammar',sub { $self->forget_grammar_click(); }],
			 );

  ## synth/say/save-wave
  $self->new_button_frame('buttons_frame',1,1,$BUTTONS_PER_ROW,'synthesis_frame','Synthesis',
			  ['stype_button','Synth Options',sub { $self->synth_options_click(); }],
			  ['synth_button','Synthesize',sub { $self->synth_click(); }],
			  ['say_button','Say',sub { $self->say_click(); }],
			  ['save_wave_button','Save Waveform',sub { $self->save_wave_click(); }],
			 );

  ## tree stuff
  $self->new_button_frame('buttons_frame',2,0,$BUTTONS_PER_ROW,'tree_frame','Tree Visualization',
			  ['show_tree_button','Show Tree',sub { $self->show_tree_click(); },],
			  ['save_tree_button','Save Tree',sub { $self->save_tree_click(); },],
			 );

  ## script stuff
  $self->new_button_frame('buttons_frame',2,1,$BUTTONS_PER_ROW,'script_frame','Scripting',
			  ['run_script_button','Run Script',sub { $self->run_script_click(); }],
			  ['save_script_button','Save As Script',sub { $self->save_script_click(); }],
			 );


  ## quit
  $tk->{quit_button} = $tk->{buttons_frame}->Button(-text => 'Quit', -relief => 'raised',
						    -command => sub { exit (0); });
  #$tk->{quit_button}->pack(-side => 'top', -fill => 'both', -expand => 1);
  $tk->{quit_button}->grid(-row=>3,-column=>0,-columnspan=>2,-sticky=>"ew");


  #------------------------------------------------------------
  # message label (statusbar)
  #------------------------------------------------------------
  $tk->{message} = '';
  my $message_lab = $tk->{message_lab} =
    $main->Label(-anchor => 'w',
		 -font => '-*-helvetica-bold-r-*-*-*-120-*-*-*-*-*-*',
		 -background => $main->cget(-background),
		 -textvariable => \$tk->{message},
		 -relief => 'groove',
		 -borderwidth => 2,
		 -width => 40
		);
  $message_lab->pack(-side => 'bottom', -fill => 'x');

  #------------------------------------------------------------
  # ye olde packlist
  #------------------------------------------------------------
  $rframe->pack(-side => 'right', -fill => 'y');
    #$buttons_frame->pack(-side => 'left', -fill => 'y');
    $rrframe->pack(-side => 'left', -fill => 'y');
      $list_label->pack(-side => 'top', -fill => 'x');
      $gen_frame->pack(-side => 'bottom', -fill => 'x');
      $listbox->pack(-side => 'left', -fill => 'both', -expand => 1);
  $lframe->pack(-side => 'top', -fill => 'both', -expand => 1);
    $xframe->pack(-side => 'top', -fill => 'both', -expand => 1);
  $buttons_frame->pack(-side => 'top', -fill => 'x');

  #------------------------------------------------------------
  # Initialization
  #------------------------------------------------------------
  #my @voices = $self->get_voice_list();
  #my ($voice);
  #foreach $voice (@voices) {
  #  my $v = $voice;
  #  $voice_sel->command(-label => $v,
	#		-command => sub { $message = "Voice set to `$v'"; $self->set_voice($v); });
  #}
  ## --- set up styles ---
  #my @styles = $self->get_style_list();
  #my ($style);
  #foreach $style (@styles) {
  #  my $s = $style;
  #  $style_sel->command(-label => $s,
  #			-command => sub { $message = "Style set to `$s'";  $self->set_style($s); });
  #}

  return $self;
}

################################################################################
# button addition
################################################################################

# $button = $self->new_button($frame_name,$row,$col,$name,$text,$command);
#   + button is grid'ed if ($row,$col) are defined, otherwise not
sub new_button {
  my ($self,$frame_name,$row,$col,$name,$text,$cmd,%args) = @_;
  my $button = $self->{tk}{$name} = $self->{tk}{$frame_name}->Button(-text => $text,
								     -relief => 'raised',
								     -command => $cmd,
								     -width => 16,
								     %args);
  $button->grid(-row=>$row,-column=>$col) if (defined($row) && defined($col));
  return $button;
}

# $frame = $self->new_button_frame($parent_name,$row,$col,$buttons_per_row,$name,$label_text);
#  + creates button-frame '$name' and label '${name}_label' in $parent_frame
#  + no label widget is created if $label_text is <undef>
#  + normally, created frame is grid()'ed into parent at ($row,$col);
#    no placement is performed if $row & $col are <undef>
sub new_button_frame {
  my ($self,$parent_name,$row,$col,$buttons_per_row,,$name,$label_text,@specs) = @_;
  my $frame = $self->{tk}{$name} = $self->{tk}{$parent_name}->Frame(-relief=>'groove',-borderwidth=>2);

  my ($brow,$bcol) = (0,0);
  if (defined($label_text)) {
    my $label = $self->{tk}{"${name}_label"} = $frame->Label(-text=>$label_text,-foreground=>'blue',-font=>'helvetica -10');
    $label->grid(-row=>0,-column=>0,-columnspan=>2,-sticky=>'n');
    ++$brow;
  }

  my ($spec);
  foreach $spec (@specs) {
    $self->new_button($name,$brow,$bcol,@$spec);
    ++$bcol;
    if ($bcol % $buttons_per_row == 0) {
      ++$brow;
      $bcol = 0;
    }
  }
  if (defined($row) && defined($col)) {
    $frame->grid(-row=>$row,-column=>$col,-sticky=>'nsew',-padx=>$padx,-pady=>$pady);
  }

  return $frame;
}


################################################################################
# Tree Canvas centering
################################################################################
sub tree_canvas_center {
  my ($canvas,$sx,$sy) = @_;
  #my $canvas = $scrolledcanvas->Subwidget('canvas');
  #my ($cx,$cy) = ($canvas->canvasx($sx), $canvas->canvasy($sy));
  #my ($cur_xvl, $cur_xvr) = $canvas->xview();
  #my ($cur_yvt, $cur_yvb) = $canvas->yview();
  #my $sr = $canvas->cget('-scrollregion');
  my ($centerx, $centery) = ( $canvas->width() / 2.0, $canvas->height() / 2.0 );
  #my $deltax = ($centerx - $cx)/10;
  #my $deltay = ($centery - $cy)/10;
  $canvas->scan('mark', $centerx, $centery);
  $canvas->scan('dragto',
		$centerx + (($centerx - $sx) / 10.0),
		$centery + (($centery - $sy) / 10.0));
}


################################################################################
# Button Functions
################################################################################


#----------------------------------------------------------
# add/del/rename
#----------------------------------------------------------
sub add_click {
  my $self = shift;
  my $id = $self->add_utterance($self->{tk}{text_field}->get(qw(0.0 end)));
  if ($id) {
    $self->{tk}{listbox}->insert('end', $id);
    $self->{tk}{listbox}->selectionClear(0,'end');
    $self->{tk}{listbox}->selectionSet('end');
    $self->{tk}{listbox}->see('end');
    $self->{tk}{message} = "Added utterance `$id'";
  } else {
    $self->{tk}{message} = 'add_utterance failed for text string.';
  }
}
sub del_click {
  my $self = shift;
  my ($uttid,$selid) = $self->get_listbox_selid();
  my $utt_tree = $self->{utts}{$uttid}{tree};
  if (defined($selid)) {
    $self->del_utterance($uttid) if (defined($uttid));
    $self->clear_display_fields();
    $self->{tk}{listbox}->selectionClear(0,'end');
    $self->{tk}{listbox}->delete($selid);
    #$tree_canvas->delete($tree_canvas->find(withtag => "$utt_tree"))
      #if (defined($utt_tree) && $utt_tree);
  } else {
    $self->{tk}{message} = "del_utterance failed for `$uttid'";
  }
}
sub rename_click {
  my $self = shift;
  my ($old_uttid,$selid) = $self->get_listbox_selid();
  my $newname = $self->{tk}{new_uttid};
  my $id = $self->rename_utterance($old_uttid,$newname);
  if ($id) {
    $self->{tk}{listbox}->delete($selid);
    $self->{tk}{listbox}->insert($selid, $id);
    $self->{tk}{listbox}->activate($selid);
    $self->{tk}{message} = "rename_utterance: Renamed `$old_uttid' to `$id'.";
  } else {
    $self->{tk}{message} = "rename_utterance from `$old_uttid' to `$newname' failed. ";
  }
}

#----------------------------------------------------------
# Grammar Stuff
#----------------------------------------------------------
sub load_grammar_click {
  my $self = shift;
  my $filename = '';

  $self->{tk}{file_selector}->configure(-title => 'Load Grammar');
  $self->set_file_entry($filename);

  $filename = $self->{tk}{file_selector}->Show;
  unless ($filename) {
    $self->{tk}{message} = "load_grammar: load operation cancelled by user.";
    return;
  }
  $self->{tk}{message} = "load_grammar: Loading `$filename'...";
  $self->update;
  $self->{tk}{main}->Busy(-recurse=>1);
  $self->load_grammar($filename);
  $self->{tk}{gen_nonterm} = $self->{grammar}->startsymbol unless ($self->{tk}{gen_nonterm});
  $self->{tk}{message} = "load_grammar: Loaded `$filename'.";
  $self->{tk}{main}->Unbusy();
}
sub retrieve_grammar_click {
  my $self = shift;
  my $filename = '';
  $self->{tk}{file_selector}->configure(-title => 'Retrieve Stored Grammar');
  $self->set_file_entry($filename);

  $filename = $self->{tk}{file_selector}->Show();
  unless ($filename) {
    $self->{tk}{message} = "retrieve_grammar: load operation cancelled by user.";
    return;
  }
  $self->{tk}{message} = "retrieve_grammar: Retrieving grammar from `$filename'...";
  $self->update;
  $self->{tk}{main}->Busy(-recurse=>1);
  $self->load_grammar($filename,'sto');
  $self->{tk}{gen_nonterm} = $self->{grammar}->startsymbol unless ($self->{tk}{gen_nonterm});
  $self->{tk}{message} = "retrieve_grammar: Grammar retrieved from `$filename'.";
  $self->{tk}{main}->Unbusy;
}
sub store_grammar_click {
  my $self = shift;
  my $grammar = $self->{grammar};
  if (defined($grammar)) {
    my $filename = '';
    $self->{tk}{file_selector}->configure(-title => 'Store Grammar');
    $self->set_file_entry($filename);

    $filename = $self->{tk}{file_selector}->Show;
    unless ($filename) {
      $self->{tk}{message} = "store_grammar: operation cancelled by user.";
      return;
    }
    $self->{tk}{message} = "store_grammar: storing grammar to file `$filename'...";
    $self->update;
    $self->{tk}{main}->Busy(-recurse=>1);
    $self->store_grammar($filename);
    $self->{tk}{message} = "store_grammar: grammar stored to file `$filename'...";
  } else {
    $self->{tk}{message} = "store_click: no grammar defined!";
  }
  $self->{tk}{main}->Unbusy();
}
sub forget_grammar_click {
  my $self = shift;
  $self->{tk}{message} = "Forgetting grammar...";
  $self->update;
  $self->{tk}{main}->Busy(-recurse=>1);
  $self->forget_grammar;
  $self->{tk}{message} = "Grammar forgotten.";
  $self->{tk}{main}->Unbusy;
}

#----------------------------------------------------------
# generate/gensay
#----------------------------------------------------------
sub generate_click {
  my $self = shift;
  my $nonterm = $self->{tk}{gen_nonterm};
  if (defined($nonterm) && $nonterm) {
    $self->{tk}{message} = "Generating derivation from `$nonterm'...";
    $self->update;
    $self->{tk}{main}->Busy(-recurse=>1);
    my $id = $self->generate($nonterm);
    if ($id) {
      $self->{tk}{message} = "Derivation from `$nonterm' generated.";
      $self->update();
      my $listbox = $self->{tk}{listbox};
      $listbox->insert('end', $id);
      $listbox->selectionClear(0,'end');
      $listbox->selectionSet('end');
      $listbox->see('end');
      $self->update();
    } else {
      $self->{tk}{message} = "generate: generation error";
    }
  } else {
    $self->{tk}{message} = "No nonterminal specified.";
    $id = '';
  }
  $self->{tk}{main}->Unbusy;
  return $id;
}
sub generate_say_click {
  my $self = shift;
  $self->generate_click();
  $self->update(); # redraw
  $self->say_click();
}

#----------------------------------------------------------
# synth-type & -options
#----------------------------------------------------------
sub synth_options_click {
  my $self = shift;
  my $tk = $self->{tk};
  my ($top);
  unless (defined($top = $self->{tk}{stype_top})) {
    $top = $self->{tk}{stype_top} = $tk->{main}->Toplevel();
    $top->title("Synthesizer Options");

    ## create radiogroup frame
    $tk->{stype_rgf} = $top->Frame(-borderwidth=>2,-relief=>'groove');
    $tk->{stype_rgf}->pack(-side=>'top',-fill=>'x',-padx=>$padx,-pady=>$pady);
    #$tk->{stype_rgf}->grid(-row=>0,-column=>0,-sticky=>'nsew',-padx=>$padx,-pady=>$pady);

    ## set radio-group variable
    $tk->{stype_var} = ref($self->{synth});
    $tk->{stype_rb_rsynth} = $tk->{stype_rgf}->Radiobutton
      (-text => 'Rsynth',
       -variable => \$tk->{stype_var},
       -value => 'SayWhat::Synthesizer::Rsynth',
       -command=> sub { $self->update_stype_opts(); },
      );
    $tk->{stype_rb_festival} = $tk->{stype_rgf}->Radiobutton
      (-text => 'Festival',
       -variable => \$tk->{stype_var},
       -value => 'SayWhat::Synthesizer::Festival',
       -command=> sub { $self->update_stype_opts(); },
      );
    $tk->{stype_rb_dummy} = $tk->{stype_rgf}->Radiobutton
      (-text => 'Dummy',
       -variable => \$tk->{stype_var},
       -value => 'SayWhat::Synthesizer::Dummy',
       -command=> sub { $self->update_stype_opts(); },
      );

    ## pack radio-group
    $tk->{stype_rb_rsynth}->grid(-row=>0,-column=>0,-sticky=>'w');
    $tk->{stype_rb_festival}->grid(-row=>1,-column=>0,-sticky=>'w');
    $tk->{stype_rb_dummy}->grid(-row=>2,-column=>0,-sticky=>'w');

    ## create options-frame
    $tk->{stype_of} = $top->Frame(-borderwidth=>2,-relief=>'groove');
    #$tk->{stype_of}->grid(-row=>1,-column=>0,-sticky=>'nsew',-padx=>$padx,-pady=>$pady);
    $tk->{stype_of}->pack(-side=>'top',-fill=>'both',-expand=>1,-padx=>$padx,-pady=>$pady);


    # add button-frame
    # + $button = $self->new_button($frame_name,$row,$col,$name,$text,$command);
    # + $frame = $self->new_button_frame($parent_name,$row,$col,$buttons_per_row,$name,$label_text);
    $self->new_button_frame('stype_top',2,0,3,'stype_bfgrid',undef,
			    ['stype_ok','OK',sub { $self->synth_options_ok(); }],
			    ['stype_apply','Apply',sub { $self->synth_options_apply(); }],
			    ['stype_cancel','Cancel',sub { $self->synth_options_cancel(); }],
			   );
    $self->{tk}{stype_bfgrid}->configure(-borderwidth=>0);
    $self->{tk}{stype_bfgrid}->pack(-side=>'bottom',-fill=>'x');

    $self->update_stype_opts();
  } else {
    $top->deiconify();
  }
}
sub synth_options_cancel {
  my $self = shift;
  $self->{tk}{stype_var} = ref($self->{synth});
  $self->{tk}{message} = "synth_type(): operation cancelled by user";
  $self->{tk}{stype_top}->withdraw;
}
sub synth_options_apply {
  my $self = shift;

  my $type = $self->synth_type($self->{tk}{stype_var});
  $self->{tk}{message} = "Set synthesizer type to '$type'";

  ## update configuration
  my $typename = $self->{tk}{stype_var};
  $typename =~ s/^.*::// if ($typename =~ /::/);
  my ($off);
  if (defined($off =  $self->{tk}{stype_of_contents}{$typename})) {
    my %config = $self->{synth}->configure();
    my ($k,$v);
    foreach $k (keys(%config)) {
      $v = $off->{$k.'_etext'};
      $v = undef if ($v eq '');
      $config{$k} = $v;
    }
    $self->{synth}->configure(%config);
  }

  $self->update_stype_opts();
}
sub synth_options_ok {
  my $self = shift;
  $self->synth_options_apply;
  $self->{tk}{stype_top}->withdraw;
}

# $tksw->update_stype_opts()
#  + updates 'synth options' frame
sub update_stype_opts {
  my $self = shift;
  my $tk = $self->{tk};
  my $tof = $tk->{stype_of};
  my $typename = ref($self->{synth});
  $typename =~ s/^.*::// if ($typename =~ /::/);
  my ($off,$sof);
  unless (defined($off = $tk->{stype_of_contents}{$typename}) &&
	  defined($sof = $off->{frame})) {
    ## create new synth-options hashref
    $off = $tk->{stype_of_contents}{$typename} = {} unless (defined($off));
    ## create new synth-options frame
    $sof = $off->{frame} = $tof->Scrolled('Frame') unless (defined($sof));
    ## create label
    $off->{label} = $sof->Label(-text=>"$typename Options");
    $off->{label}->grid(-row=>0,-column=>0,-columnspan=>2,-sticky=>'ew');
    ## create options
    my %config = $self->{synth}->configure();
    my ($k,$v);
    my $row = 1;
    foreach $k (sort(keys(%config))) {
      ## option label
      $off->{$k.'_l'} = $sof->Label(-text=>$k,-anchor=>'e');
      $off->{$k.'_l'}->grid(-row=>$row,-column=>0,-sticky=>'e');
      ## option entry
      $off->{$k.'_etext'} = defined($config{$k}) ? $config{$k} : '';
      $off->{$k.'_e'} = $sof->Entry(-textvariable=>\$off->{$k.'_etext'});
      $off->{$k.'_e'}->grid(-row=>$row,-column=>1,-sticky=>'ew');
      ++$row;
    }
  }
  $tk->{stype_of_current}->packForget() if (defined($tk->{stype_of_current}));
  $sof->pack(-side=>'top',-fill=>'both',-expand=>'true',-padx=>$padx,-pady=>$pady);
  $tk->{stype_of_current} = $sof;
}

#----------------------------------------------------------
# synth/say/save-wave # NOTE: $w->Busy(recurse=>1); $w->Unbusy();
#----------------------------------------------------------
sub synth_click {
  my $self = shift;
  my ($uttid, $selid) = $self->get_listbox_selid();
  if (defined($uttid)) {
    $self->{tk}{message} = "Synthesizing utterance `$uttid'...";
    $self->synth_utterance($uttid);
    $self->{tk}{message} = "Synthesis of utterance `$uttid' complete.";
  } else {
    $self->{tk}{message} = "synth_utterance: no utterance selected.";
  }
}
sub say_click {
  my $self = shift;
  my ($uttid, $selid) = $self->get_listbox_selid();
  if (defined($uttid)) {
    $self->{tk}{message} = "say_utterance: saying utterance `$uttid'";
    $self->update();
    $self->{tk}{main}->Busy(-recurse=>1);
    $self->say_utterance($uttid);
    $self->{tk}{message} = "say_utterance: pronunciation of `$uttid' complete";
    $self->{tk}{main}->Unbusy;
    $self->update();
  } else {
    $self->{tk}{message} = "say_utterance: no utterance selected.";
  }
}
sub save_wave_click {
  my $self = shift;
  my ($uttid, $selid) = $self->get_listbox_selid();
  if (defined($uttid)) {
    my $filename = "$uttid.$self->{synth}{mode}";
    $self->{tk}{file_selector}->configure(-title => 'Save Utterance Waveform');
    $self->set_file_entry($filename);

    $filename = $self->{tk}{file_selector}->Show();
    unless ($filename) {
      $message = "save_utterance: operation cancelled by user.";
      return;
    }
    $self->{tk}{message} = "save_utterance: saving utterance `$uttid' to file `$filename'...";
    $self->save_utterance($uttid,$filename);
    $self->{tk}{message} = "save_utterance: Utterance `$uttid' saved to file `$filename'.";
  } else {
    $self->{tk}{message} = "save_utterance: no utterance selected.";
  }
}


#----------------------------------------------------------
# Listbox Stuff
#----------------------------------------------------------
sub update {
  my $self = shift;
  my ($uttid,$selid) = $self->get_listbox_selid();
  if (defined($selid)) {
    $self->{tk}{new_uttid} = $uttid;
    $self->{tk}{text_field}->delete('0.0', 'end');
    $self->{tk}{text_field}->insert('0.0', $self->{utts}{$uttid}{text});
  }
  # redraw
  $self->{tk}{main}->update();
}

#----------------------------------------------------------
# Tree stuff
# --> NEEDS FIXING!
#----------------------------------------------------------
sub show_tree_click {
  my $self = shift;
  my $tk = $self->{tk};
  $tk->{main}->Busy(-recurse=>1);

  my ($uttid,$selid) = $self->get_listbox_selid();
  if ($uttid) {
    my ($str,$tree) = @{$self->{utts}{$uttid}}{qw(text tree)};

    ## setup tree toplevel
    my $top = $tk->{tree_top};
    my $tree_canvas = $tk->{tree_canvas};
    unless (defined($top)) {
      ## create if not defined
      $top = $tk->{tree_top} = $tk->{main}->Toplevel();

      $tree_canvas = $tk->{tree_canvas} = $top->Scrolled('Canvas',
							 -relief => 'sunken',
							 -cursor => 'target',
							 -borderwidth => 2,
							 -background => 'white',
							 -scrollbars => 'se',
							 -confine => 1);
      $tk->{tree_canvas}->pack(-side => 'top', -fill => 'both', -expand => 1);

      ## setup canvas bindings
      $tk->{tree_canvas}->CanvasBind
	("<Button-2>",[$tk->{tree_canvas}, 'scan', 'mark', Ev('x'), Ev('y')] );
      $tk->{tree_canvas}->CanvasBind
	("<Button2-Motion>", [$tk->{tree_canvas}, 'scan', 'dragto', Ev('x'), Ev('y')] );
      $tk->{tree_canvas}->CanvasBind
	("<Button-1>", [\&tree_canvas_center, Ev('x'), Ev('y')] );

      # $button = $self->new_button($parent_name,$row,$col,$name,$text,$command);
      $self->new_button('tree_top',undef,undef,'tree_dismiss','Dismiss',sub { $self->tree_dismiss_click(); },);
      $tk->{tree_dismiss}->pack(-side=>'top',-fill=>'x');
    }

    ## show top-level
    $top->title("Derivation Tree: '$uttid'");
    $top->deiconify();

    ## clear tree-canvas
    $tree_canvas->delete('all');

    if (defined($tree) && $tree) {
      ## draw tree
      my @tree_items = qw();
      # gotta generate it
      draw_tree($tree,$tree_canvas);
    }
    $self->{tk}{tree_is_showing} = 1;
  }
  else {
    $message = "show_utterance: no utterance selected.";
  }
  $tk->{main}->Unbusy();
}
sub tree_dismiss_click {
  my $self = shift;
  $self->{tk}{tree_top}->withdraw();
  $self->{tk}{tree_is_showing} = 0;
}

sub save_tree_click {
  my $self = shift;
  $self->show_tree_click() unless ($self->{tk}{tree_is_showing});
  my ($uttid, $selid) = $self->get_listbox_selid();
  if (defined($uttid)) {
    my $filename = "$uttid.ps";
    my $file_selector = $self->{tk}{file_selector};

    $file_selector->configure(-title => 'Save Utterance Tree Postscript');
    $self->set_file_entry($filename);

    $filename = $file_selector->Show();
    unless ($filename) {
      $self->{tk}{message} = "save_tree: operation cancelled by user.";
      return;
    }
    $filename .= '.ps' unless ($filename =~ /\.ps$/);
    $self->{tk}{message} = "save_tree: saving tree for `$uttid' as postscript to file `$filename'...";
    $self->update();
    # actual save
    my $bb = $self->{tk}{tree_canvas}->cget('-scrollregion');
    $self->{tk}{tree_canvas}->postscript(
					 -file => $filename,
					 -height => $bb->[3] - $bb->[1],
					 -width => $bb->[2] - $bb->[0],
					 -x => $bb->[0],
					 -y => $bb->[1],
					);
    # end actual save
    $self->{tk}{message} = "save_utterance: tree for `$uttid' saved as postscript to file `$filename'.";
  } else {
    $self->{tk}{message} = "save_utterance: no utterance selected.";
  }
}




#----------------------------------------------------------
# script stuff
#----------------------------------------------------------
sub run_script_click {
  my $self = shift;
  my $filename = '';
  $self->{tk}{file_selector}->configure(-title => 'Load SayWhat Script');
  $self->set_file_entry($filename);

  $filename = $self->{tk}{file_selector}->Show;
  unless ($filename) {
    $self->{tk}{message} = "run_script: load operation cancelled by user.";
    return;
  }
  unless (open(INFILE, "<$filename")) {
    $self->{tk}{message} = "run_script: open failed for `$filename': $!";
    return undef;
  }
  ## actual load
  $self->{tk}{message} = "Running script `$filename'";
  $self->update;
  $self->{tk}{main}->Busy(-recurse=>1);
  #$main::lexer->from(*INFILE);
  $main::lexwrapper->from(*INFILE);
  $main::lexwrapper->filename($filename);
  main::doparse();
  close(INFILE);
  $self->{tk}{message} = "run_script: loaded `$filename'.";
  $self->{tk}{main}->Unbusy();
  $self->refresh_listbox();
}
sub save_script_click {
  my $self = shift;
  my $filename = '';
  $self->{tk}{file_selector}->configure(-title => 'Save SayWhat Script');
  $self->set_file_entry($filename);

  $filename = $file_selector->Show;
  unless ($filename) {
    $self->{tk}{message} = "save_script: operation cancelled by user.";
    return;
  }
  $self->{tk}{message} = "save_script: saving state to script file `$filename'...";
  unless ($self->save_script($filename)) {
    $self->{tk}{message} = "save_script: save to `$filename' failed: $!";
  } else {
    $self->{tk}{message} = "State saved to script file `$filename'.";
  }
}


#----------------------------------------------------------
# Utilities
#----------------------------------------------------------
sub get_listbox_selid {
  my $self = shift;
  my ($idx) = $self->{tk}{listbox}->curselection();
  return undef unless (defined($idx) && $idx ne '');
  return ($self->{tk}{listbox}->get($idx), $idx);
}

sub clear_display_fields {
  my $self = shift;
  $self->{tk}{text_field}->delete('0.0','end');
  ##$tree_canvas->delete($tree_canvas->find('all'));
  #$self->{tk}{tree_canvas}->itemconfigure('all', -fill => 'white');
}

sub refresh_listbox {
  my $self = shift;
  $self->{tk}{listbox}->delete('0','end');
  my ($id);
  foreach $id (sort(keys(%{$self->{utts}}))) {
    $self->{tk}{listbox}->insert('end', $id);
  }
}

sub file_entry_return_click {
  my $fs = shift;
  $fs->Subwidget('file_entry')->selectionRange(0,'end');
  $fs->Subwidget('file_entry')->bind('<Key-Return>', sub { file_entry_return_click($fs); });
  $fs->Accept;
}

sub set_file_entry {
  my ($self,$filename) = @_;
  my $entry = $self->{tk}{file_selector}->Subwidget('file_entry');
  $entry->delete(0,'end');
  $entry->insert(0,$filename);
  $entry->selectionRange(0,'end');
}

##########################################################################
# Festival communication
##########################################################################
sub get_voice_list {
  my $self = shift;
  return Speech::Festival::Synthesiser::voice_list($self->{festival});
}
sub get_style_list { # hacked from Speech::Festival::Synthesiser::voice_list
  my $self = shift;
  my (@res) = ();

  $self->{festival}->request("(mapcar car Styles)", \&Speech::Festival::Synthesiser::ret_lp_handler, \@res);
  my ($scheme) = Speech::Festival::parse_scheme($res[0]);
  # @res = split(/\s+/, ($res[0] =~ /\(\s*(.*?)\s*\)/)[0]);
  
  return @$scheme;
}

##########################################################################
# Tree-drawing
##########################################################################
sub draw_tree {
  my ($tree,$canvas) = @_;
  return unless (defined($tree) && defined($canvas));
  $tree->draw_tree_tk(canvas => $canvas,
		      xskip => $Tree_SkipX,
		      yskip => $Tree_SkipY,
		      linepad => $Tree_LinePad,
		      font => $Tree_Font,
		      boxactive => $Tree_BoxActive);
}

1;
